import json

import pytest
from requests.structures import CaseInsensitiveDict
from petisco.webhooks.webhook.domain.webhook_response_result import (
    WebhookResponseResult,
)


@pytest.mark.unit
def test_should_construct_and_convert_to_dict_webhook_response_result_from_case_insensitive_dict_headers():

    expected_headers = {
        "X-Petisco-Event": "user_created",
        "X-Petisco-Event-Version": "1",
        "X-Petisco-Delivery": "a026531a-8512-4599-9491-81909ea8e31b",
        "X-Petisco-WebhookId": "1baa14dc-bda9-400f-b45b-428f26532f1d",
        "X-Petisco-Request-Timestamp": "2020-11-18 11:38:37.156162",
        "User-Agent": "Petisco-Hookshoot/",
        "apikey": "'b0b905d6-228f-44bf-a130-c85d7aecd765",
    }
    headers = CaseInsensitiveDict(expected_headers)
    body = {"payload": "ok"}
    status_code = 200
    webhook_response_result = WebhookResponseResult(
        headers=headers, body=body, status_code=status_code
    )

    assert webhook_response_result.to_dict() == {
        "headers": json.dumps(expected_headers),
        "body": json.dumps(body),
        "status_code": status_code,
        "completed_in_ms": None,
    }


@pytest.mark.unit
def test_should_construct_and_convert_to_dict_webhook_response_result_from_default_dicts():

    headers = {
        "X-Petisco-Event": "user_created",
        "X-Petisco-Event-Version": "1",
        "X-Petisco-Delivery": "a026531a-8512-4599-9491-81909ea8e31b",
        "X-Petisco-WebhookId": "1baa14dc-bda9-400f-b45b-428f26532f1d",
        "X-Petisco-Request-Timestamp": "2020-11-18 11:38:37.156162",
        "User-Agent": "Petisco-Hookshoot/",
        "apikey": "'b0b905d6-228f-44bf-a130-c85d7aecd765",
    }
    body = {"payload": "ok"}
    status_code = 200
    completed_in_ms = 80

    webhook_response_result = WebhookResponseResult(
        headers=headers,
        body=body,
        status_code=status_code,
        completed_in_ms=completed_in_ms,
    )

    assert webhook_response_result.to_dict() == {
        "headers": json.dumps(headers),
        "body": json.dumps(body),
        "status_code": status_code,
        "completed_in_ms": completed_in_ms,
    }



import copy
from typing import Dict
from abc import ABCMeta, abstractmethod

from petisco.domain.aggregate_roots.info_id import InfoId
from petisco.event.shared.domain.event import Event, Events


class IEventBus:

    __metaclass__ = ABCMeta

    @classmethod
    def __repr__(cls):
        return f"{cls.__name__}"

    @classmethod
    def info(cls) -> Dict:
        return {"name": cls.__name__}

    def _set_info_id(self, info_id: InfoId):
        self.info_id = info_id

    def _set_additional_meta(self, meta: Dict):
        self.additional_meta = meta

    def with_info_id(self, info_id: InfoId):
        event_bus = copy.copy(self)
        event_bus._set_info_id(info_id)
        return event_bus

    def with_meta(self, meta: Dict, info_id: InfoId = None):
        event_bus = copy.copy(self)
        event_bus._set_additional_meta(meta)
        if info_id:
            event_bus._set_info_id(info_id)
        return event_bus

    @abstractmethod
    def publish(self, event: Event):
        raise NotImplementedError

    def publish_events(self, events: Events):
        for event in events:
            self.publish(event)

    @abstractmethod
    def retry_publish_only_on_store_queue(self, event: Event):
        raise NotImplementedError

    def retry_publish_events_only_on_store_queue(self, events: Events):
        for event in events:
            self.retry_publish_only_on_store_queue(event)

    @abstractmethod
    def stop(self):
        raise NotImplementedError



import pytest

from petisco.domain.value_objects.string_value_object import (
    StringValueObject,
    InvalidStringValueObjectError,
)
from petisco.domain.value_objects.value_object import ValueObjectError


@pytest.mark.unit
def test_should_string_value_object_construct():
    value = "acme"
    vo = StringValueObject(value)

    assert vo.value == value


@pytest.mark.unit
def test_should_inherit_from_string_value_object_and_construct():
    class Name(StringValueObject):
        pass

    value = "acme"

    name = Name(value)

    assert isinstance(name, Name)
    assert name.value == value


@pytest.mark.unit
def test_should_inherit_from_string_value_object_and_add_an_ensure_clause():
    class Name(StringValueObject):
        def __init__(self, value: str):
            super(Name, self).__init__(value)

        def guard(self):
            self._ensure_name_is_less_than_30_chars()

        def _ensure_name_is_less_than_30_chars(self):
            if len(self.value) >= 30:
                raise ValueObjectError()

    value = "acme"
    name = Name(value)

    assert isinstance(name, Name)
    assert name.value == value

    invalid_value = "a" * 35
    with pytest.raises(ValueObjectError):
        Name(invalid_value)


@pytest.mark.unit
@pytest.mark.parametrize("value", ["Alex", "Íñigo", "João"])
def test_should_inherit_from_string_value_object_with_disallowed_utf8mb4_and_create_objects_successfully(
    value,
):
    class Name(StringValueObject):
        def __init__(self, value: str):
            super(Name, self).__init__(value)

        def guard(self):
            self._ensure_value_contains_valid_char(allow_utf8mb4=False)

    name = Name(value)
    assert isinstance(name, Name)
    assert name.value == value


@pytest.mark.unit
@pytest.mark.parametrize("value", ["𝘼𝙡𝙚𝙭", "Alex𖥔", "Alex😃"])
def test_should_inherit_from_string_value_object_with_disallowed_utf8mb4_and_raise_exception_with_utf8mb4_values(
    value,
):
    class Name(StringValueObject):
        def __init__(self, value: str):
            super(Name, self).__init__(value)

        def guard(self):
            self._ensure_value_contains_valid_char(allow_utf8mb4=False)

    with pytest.raises(InvalidStringValueObjectError):
        Name(value)



from petisco.tasks.config.config_tasks import ConfigTasks
from petisco.tasks.domain.interface_task_executor import TaskExecutor


class NotImplementedTaskExecutor(TaskExecutor):
    def start(self, config_tasks: ConfigTasks):
        pass

    def stop(self):
        pass



from abc import ABCMeta, abstractmethod


class BaseObject:

    __metaclass__ = ABCMeta

    def __init__(self, execute_guard: bool = True):
        if execute_guard:
            self.guard()

    @abstractmethod
    def guard(self):
        pass



from petisco import INFO, event_from_redis_message
from petisco.application.petisco import Petisco


def redis_event_handler(message):
    event = event_from_redis_message(message).unwrap_or_return()
    logger = Petisco.get_instance()._logger
    logger.log(INFO, event)



import json
from json.decoder import JSONDecodeError
from typing import Tuple

from meiga import Error

from petisco.domain.timedelta_parser import TimeDeltaParser


def get_error_message(error_message_title, response):
    error_message = {"error": error_message_title}

    try:
        detail = response.json()
        error_message["detail"] = detail
    except JSONDecodeError:
        pass

    return error_message


class RequestError(Error):
    def __init__(
        self,
        error_name: str,
        error_message: str,
        status_code: int,
        url: str,
        headers: dict = None,
        content: dict = None,
        completed_in_ms: float = None,
    ):
        self.error_name = error_name
        self.error_message = error_message
        self.status_code = status_code
        self.url = url
        self.headers = headers if headers else None
        self._set_content(content)
        self.completed_in_ms = completed_in_ms

    def _set_content(self, content):
        self.content = None
        if content:
            try:
                if isinstance(content, bytes):
                    content = content.decode()
                self.content = json.loads(content)
            except Exception:
                pass

    def __repr__(self):
        return f"{self.error_name} ({self.status_code} | {self.url}) -> {self.error_message} (completed_in: {self.completed_in_ms})"


class MultipartFormatRequestError(RequestError):
    @staticmethod
    def from_response(response):
        error_message = get_error_message("Fail to create multipart", response)
        return MultipartFormatRequestError(
            error_name="MultipartFormatRequestError",
            error_message=json.dumps(error_message),
            status_code=response.status_code,
            url=response.url,
            headers=response.headers,
            content=response.content,
            completed_in_ms=TimeDeltaParser.ms_from_timedelta(response.elapsed),
        )


class MissingSchemaRequestError(RequestError):
    def __init__(self, url: str):
        super().__init__(
            error_name="MissingSchemaRequestError",
            error_message=json.dumps({"error": "Missing schema in request"}),
            status_code=422,
            url=url,
        )


class TimeoutRequestError(RequestError):
    def __init__(self, url: str, timeout: Tuple[int, int]):
        super().__init__(
            error_name="TimeoutRequestError",
            error_message=json.dumps({"error": "Timeout error"}),
            status_code=408,
            url=url,
        )
        self.timeout = timeout


class ConnectionRequestError(RequestError):
    def __init__(self, url: str):
        super().__init__(
            error_name="ConnectionRequestError",
            error_message=json.dumps({"error": "Connection error"}),
            status_code=503,
            url=url,
        )


class BadRequestError(RequestError):
    @staticmethod
    def from_response(response):
        error_message = get_error_message("Bad request Error", response)
        return BadRequestError(
            error_name="BadRequestError",
            error_message=json.dumps(error_message),
            status_code=response.status_code,
            url=response.url,
            headers=response.headers,
            content=response.content if response.content else response.text,
            completed_in_ms=TimeDeltaParser.ms_from_timedelta(response.elapsed),
        )


class UnauthorizedRequestError(RequestError):
    @staticmethod
    def from_response(response):
        error_message = get_error_message("Unauthorized error", response)
        return UnauthorizedRequestError(
            error_name="UnauthorizedRequestError",
            error_message=json.dumps(error_message),
            status_code=response.status_code,
            url=response.url,
            headers=response.headers,
            content=response.content,
            completed_in_ms=TimeDeltaParser.ms_from_timedelta(response.elapsed),
        )


class UnknownRequestError(RequestError):
    @staticmethod
    def from_response(response):
        error_message = get_error_message("Unknown error", response)
        return UnknownRequestError(
            error_name="UnknownRequestError",
            error_message=json.dumps(error_message),
            status_code=response.status_code,
            url=response.url,
            headers=response.headers,
            content=response.content,
            completed_in_ms=TimeDeltaParser.ms_from_timedelta(response.elapsed),
        )

    @staticmethod
    def from_exception(exc, url: str):
        return UnknownRequestError(
            error_name="UnknownRequestError",
            error_message=str(exc),
            status_code=500,
            url=url,
        )



from tests.modules.event.mothers.defaults import DEFAULT_SERVICE


class RabbitMqQueueNamingMother:
    @staticmethod
    def legacy_main_queue():
        return f"{DEFAULT_SERVICE}-event"

    @staticmethod
    def legacy_dead_letter_queue():
        return f"dl-{RabbitMqQueueNamingMother.legacy_main_queue()}"



from sqlalchemy import Column, Integer

from petisco.persistence.persistence import Persistence

Base = Persistence.get_base("petisco-sql")


class UsersCountModel(Base):
    __tablename__ = "UsersCount"

    count = Column("count", Integer, primary_key=True)



import pytest
from meiga import Success, Result, Error, isFailure

from petisco import Executable, PoolExecutor


@pytest.mark.unit
@pytest.mark.parametrize(
    "use_multi_threading,unwrap_results,expected_result",
    [
        (True, True, [3, 4, 2, 10]),
        (False, True, [3, 4, 2, 10]),
        (True, False, [Success(3), Success(4), Success(2), Success(10)]),
        (False, False, [Success(3), Success(4), Success(2), Success(10)]),
    ],
)
def test_pool_executor_should_execute_several_executables(
    use_multi_threading, unwrap_results, expected_result
):
    def sum(param_1, param_2) -> Result[int, Error]:
        return Success(param_1 + param_2)

    def mul(param_1, param_2) -> Result[int, Error]:
        return Success(param_1 * param_2)

    executables = [
        Executable(func=sum, args=(1, 2)),
        Executable(func=sum, args=(2, 2)),
        Executable(func=mul, args=(1, 2)),
        Executable(func=mul, args=(5, 2)),
    ]

    executor = PoolExecutor(executables, use_multi_threading)

    results = executor.execute()

    if unwrap_results:
        results = [result.unwrap_or_return() for result in results]

    assert results == expected_result


@pytest.mark.unit
@pytest.mark.parametrize("use_multi_threading", [True, False])
def test_pool_executor_should_fail_when_executable_returns_a_failure(
    use_multi_threading,
):
    def sum(param_1, param_2) -> Result[int, Error]:
        return isFailure

    def mul(param_1, param_2) -> Result[int, Error]:
        return Success(param_1 * param_2)

    executables = [Executable(func=sum, args=(1, 2)), Executable(func=mul, args=(1, 2))]

    executor = PoolExecutor(executables, use_multi_threading)

    results = executor.execute()

    assert results[0].is_failure
    assert results[1].is_success



import pytest
from meiga import Success
from meiga.assertions import assert_failure, assert_success
from meiga.decorators import meiga

from petisco.domain.value_objects.user_id import UserId
from petisco.domain.value_objects.uuid import InvalidUuidError


@pytest.mark.unit
@pytest.mark.parametrize(
    "value", ["0VAnhfnTtMJk3SBN", "9f74b5c0-2196-4a27-a03f-cb92d66e2bbd"]
)
def test_should_declare_a_user_id_manually(value):

    user_id = UserId(value)

    assert isinstance(user_id, UserId)
    assert user_id.value == value


@pytest.mark.unit
@pytest.mark.parametrize(
    "value", ["user_id", "my_user_id_is_too_long_for_default_limit_length"]
)
def test_should_fail_user_id_when_declare_a_non_valid_user_id(value):
    @meiga
    def controller():
        user_id = UserId(value)
        return Success(user_id)

    result = controller()
    assert_failure(result, value_is_instance_of=InvalidUuidError)


@pytest.mark.unit
def test_should_declare_a_user_id_with_empty_string():
    with pytest.raises(InvalidUuidError):
        UserId("")


@pytest.mark.unit
def test_should_declare_a_user_id_with_none():
    with pytest.raises(InvalidUuidError):
        UserId(None)


@pytest.mark.unit
def test_should_declare_a_user_id_generated_automatically():
    user_id = UserId.generate()

    assert isinstance(user_id, UserId)
    assert len(user_id.value) == 36


@pytest.mark.unit
def test_should_declare_a_legacy_user_id_generated_automatically():
    user_id = UserId.generate_legacy()

    assert isinstance(user_id, UserId)
    assert len(user_id.value) == 16


@pytest.mark.unit
def test_should_fail_when_declare_a_user_id_that_exceeds_default_length_limits():
    with pytest.raises(InvalidUuidError):
        UserId("my_user_id_is_too_long_for_default_limit_length")


@pytest.mark.unit
def test_should_return_success_user_id_if_generate_user_id_on_meiga_decorated_method():
    @meiga
    def controller():
        return Success(UserId.generate())

    result = controller()
    assert_success(result, value_is_instance_of=UserId)
